@@ -632,8 +632,8 @@ static void
 test_ubandsel_response_four (void)
 {
     const MMModemBand expected_bands[] = {
-        /*  850 */ MM_MODEM_BAND_G850, MM_MODEM_BAND_U850, MM_MODEM_BAND_EUTRAN_5,
-        /*  900 */ MM_MODEM_BAND_EGSM, MM_MODEM_BAND_U900, MM_MODEM_BAND_EUTRAN_8,
+        /*  850 */ MM_MODEM_BAND_G850, MM_MODEM_BAND_UTRAN_5, MM_MODEM_BAND_EUTRAN_5,
+        /*  900 */ MM_MODEM_BAND_EGSM, MM_MODEM_BAND_UTRAN_8, MM_MODEM_BAND_EUTRAN_8,
         /* 1800 */ MM_MODEM_BAND_DCS, MM_MODEM_BAND_UTRAN_3, MM_MODEM_BAND_EUTRAN_3,
         /* 1900 */ MM_MODEM_BAND_PCS, MM_MODEM_BAND_UTRAN_2, MM_MODEM_BAND_EUTRAN_2,
     };
@@ -741,14 +741,22 @@ common_validate_uact_response (const gchar       *str,
     GArray *bands;
 
     bands = mm_ublox_parse_uact_response (str, &error);
-    g_assert_no_error (error);
 
-    common_compare_bands (bands, expected_bands, n_expected_bands);
+    if (n_expected_bands > 0) {
+        g_assert (bands);
+        g_assert_no_error (error);
+        common_compare_bands (bands, expected_bands, n_expected_bands);
+    } else {
+        g_assert (!bands);
+        g_assert (error);
+        g_error_free (error);
+    }
 }
 
 static void
 test_uact_response_empty_list (void)
 {
+    common_validate_uact_response ("", NULL, 0);
     common_validate_uact_response ("+UACT: ,,,\r\n", NULL, 0);
 }
 
