@@ -596,7 +596,7 @@ build_start_network_input (ConnectContext *ctx)
 
     /* Need to add auth info? */
     if (has_user || has_password || ctx->auth != QMI_WDS_AUTHENTICATION_NONE) {
-        /* We define a valid auth preference if we have either user or password, or a explicit
+        /* We define a valid auth preference if we have either user or password, or an explicit
          * request for one to be set. If no explicit one was given, default to PAP. */
         qmi_message_wds_start_network_input_set_authentication_preference (
             input,
@@ -1851,9 +1851,10 @@ disconnect_context_step (GTask *task)
                                                                            ctx->client_ipv4,
                                                                            FALSE,
                                                                            &self->priv->packet_service_status_ipv4_indication_id);
-            cleanup_event_report_unsolicited_events (self,
-                                                     ctx->client_ipv4,
-                                                     &self->priv->event_report_ipv4_indication_id);
+            if (self->priv->event_report_ipv4_indication_id)
+                cleanup_event_report_unsolicited_events (self,
+                                                         ctx->client_ipv4,
+                                                         &self->priv->event_report_ipv4_indication_id);
 
             input = qmi_message_wds_stop_network_input_new ();
             qmi_message_wds_stop_network_input_set_packet_data_handle (input, ctx->packet_data_handle_ipv4, NULL);
@@ -1880,9 +1881,10 @@ disconnect_context_step (GTask *task)
                                                                            ctx->client_ipv6,
                                                                            FALSE,
                                                                            &self->priv->packet_service_status_ipv6_indication_id);
-            cleanup_event_report_unsolicited_events (self,
-                                                     ctx->client_ipv6,
-                                                     &self->priv->event_report_ipv6_indication_id);
+            if (self->priv->event_report_ipv6_indication_id)
+                cleanup_event_report_unsolicited_events (self,
+                                                         ctx->client_ipv6,
+                                                         &self->priv->event_report_ipv6_indication_id);
 
             input = qmi_message_wds_stop_network_input_new ();
             qmi_message_wds_stop_network_input_set_packet_data_handle (input, ctx->packet_data_handle_ipv6, NULL);
