@@ -28,11 +28,44 @@
 #include "mm-iface-modem-simple.h"
 #include "mm-log.h"
 
+/*****************************************************************************/
+/* Private data context */
+
+#define PRIVATE_TAG "iface-modem-simple-private-tag"
+static GQuark private_quark;
+
+typedef struct {
+    GCancellable *ongoing_connect;
+} Private;
+
+static void
+private_free (Private *priv)
+{
+    g_assert (!priv->ongoing_connect);
+    g_slice_free (Private, priv);
+}
+
+static Private *
+get_private (MMIfaceModemSimple *self)
+{
+    Private *priv;
+
+    if (G_UNLIKELY (!private_quark))
+        private_quark = g_quark_from_static_string (PRIVATE_TAG);
+
+    priv = g_object_get_qdata (G_OBJECT (self), private_quark);
+    if (!priv) {
+        priv = g_slice_new0 (Private);
+        g_object_set_qdata_full (G_OBJECT (self), private_quark, priv, (GDestroyNotify)private_free);
+    }
+
+    return priv;
+}
+
 /*****************************************************************************/
 /* Register in either a CDMA or a 3GPP network (or both) */
 
 typedef struct {
-    MMIfaceModemSimple *self;
     gchar *operator_id;
     guint remaining_tries_cdma;
     guint remaining_tries_3gpp;
@@ -43,7 +76,6 @@ static void
 register_in_network_context_free (RegisterInNetworkContext *ctx)
 {
     g_free (ctx->operator_id);
-    g_object_unref (ctx->self);
     g_free (ctx);
 }
 
@@ -67,8 +99,7 @@ register_in_cdma_network_ready (MMIfaceModemCdma *self,
     ctx = g_task_get_task_data (task);
     ctx->remaining_tries_cdma--;
 
-    if (!mm_iface_modem_cdma_register_in_network_finish (
-            MM_IFACE_MODEM_CDMA (self), res, NULL)) {
+    if (!mm_iface_modem_cdma_register_in_network_finish (self, res, NULL)) {
         /* Retry check */
         check_next_registration (task);
         return;
@@ -89,8 +120,7 @@ register_in_3gpp_network_ready (MMIfaceModem3gpp *self,
     ctx = g_task_get_task_data (task);
     ctx->remaining_tries_3gpp--;
 
-    if (!mm_iface_modem_3gpp_register_in_network_finish (
-            MM_IFACE_MODEM_3GPP (self), res, NULL)) {
+    if (!mm_iface_modem_3gpp_register_in_network_finish (self, res, NULL)) {
         /* Retry check */
         check_next_registration (task);
         return;
@@ -105,13 +135,15 @@ static void
 check_next_registration (GTask *task)
 {
     RegisterInNetworkContext *ctx;
+    MMIfaceModemSimple *self;
 
+    self = MM_IFACE_MODEM_SIMPLE (g_task_get_source_object (task));
     ctx = g_task_get_task_data (task);
 
     if (ctx->remaining_tries_cdma > ctx->remaining_tries_3gpp &&
         ctx->remaining_tries_cdma > 0) {
         mm_iface_modem_cdma_register_in_network (
-            MM_IFACE_MODEM_CDMA (ctx->self),
+            MM_IFACE_MODEM_CDMA (self),
             ctx->max_try_time,
             (GAsyncReadyCallback)register_in_cdma_network_ready,
             task);
@@ -120,7 +152,7 @@ check_next_registration (GTask *task)
 
     if (ctx->remaining_tries_3gpp > 0) {
         mm_iface_modem_3gpp_register_in_network (
-            MM_IFACE_MODEM_3GPP (ctx->self),
+            MM_IFACE_MODEM_3GPP (self),
             ctx->operator_id,
             ctx->max_try_time,
             (GAsyncReadyCallback)register_in_3gpp_network_ready,
@@ -145,17 +177,16 @@ register_in_3gpp_or_cdma_network (MMIfaceModemSimple *self,
     GTask *task;
 
     ctx = g_new0 (RegisterInNetworkContext, 1);
-    ctx->self = g_object_ref (self);
     ctx->operator_id = g_strdup (operator_id);
 
     /* 3GPP-only modems... */
-    if (mm_iface_modem_is_3gpp_only (MM_IFACE_MODEM (ctx->self))) {
+    if (mm_iface_modem_is_3gpp_only (MM_IFACE_MODEM (self))) {
         ctx->max_try_time = 60;
         ctx->remaining_tries_cdma = 0;
         ctx->remaining_tries_3gpp = 1;
     }
     /* CDMA-only modems... */
-    else if (mm_iface_modem_is_cdma_only (MM_IFACE_MODEM (ctx->self))) {
+    else if (mm_iface_modem_is_cdma_only (MM_IFACE_MODEM (self))) {
         ctx->max_try_time = 60;
         ctx->remaining_tries_cdma = 1;
         ctx->remaining_tries_3gpp = 0;
@@ -201,16 +232,18 @@ typedef struct {
 
     /* Results to set */
     MMBaseBearer *bearer;
+
+    /* Cancellation */
+    GCancellable *cancellable;
 } ConnectionContext;
 
 static void
 connection_context_free (ConnectionContext *ctx)
 {
+    g_clear_object (&ctx->cancellable);
+    g_clear_object (&ctx->properties);
+    g_clear_object (&ctx->bearer);
     g_variant_unref (ctx->dictionary);
-    if (ctx->properties)
-        g_object_unref (ctx->properties);
-    if (ctx->bearer)
-        g_object_unref (ctx->bearer);
     g_object_unref (ctx->skeleton);
     g_object_unref (ctx->invocation);
     g_object_unref (ctx->self);
@@ -439,9 +472,58 @@ bearer_list_find_disconnected (MMBaseBearer *bearer,
         ctx->found = g_object_ref (bearer);
 }
 
+static gboolean
+completed_if_cancelled (ConnectionContext *ctx)
+{
+    /* Do nothing if not cancelled */
+    if (!g_cancellable_is_cancelled (ctx->cancellable))
+        return FALSE;
+
+    /* Otherwise cancellable is valid and it's cancelled */
+    g_dbus_method_invocation_return_error (
+        ctx->invocation, G_IO_ERROR, G_IO_ERROR_CANCELLED,
+        "Connection attempt cancelled");
+    connection_context_free (ctx);
+    return TRUE;
+}
+
+static void
+cleanup_cancellation (ConnectionContext *ctx)
+{
+    Private *priv;
+
+    priv = get_private (ctx->self);
+    g_clear_object (&priv->ongoing_connect);
+    g_clear_object (&ctx->cancellable);
+}
+
+static gboolean
+setup_cancellation (ConnectionContext  *ctx,
+                    GError            **error)
+{
+    Private *priv;
+
+    /* Only one connection attempt by Simple.Connect() may be handled at a time */
+    priv = get_private (ctx->self);
+    if (priv->ongoing_connect) {
+        g_set_error (error, MM_CORE_ERROR, MM_CORE_ERROR_IN_PROGRESS,
+                     "Connection request forbidden: operation already in progress");
+        return FALSE;
+    }
+
+    g_assert (!ctx->cancellable);
+    ctx->cancellable = g_cancellable_new ();
+    priv->ongoing_connect = g_object_ref (ctx->cancellable);
+    return TRUE;
+}
+
 static void
 connection_step (ConnectionContext *ctx)
 {
+    /* Early abort if operation is cancelled */
+    if (completed_if_cancelled (ctx))
+        return;
+
     switch (ctx->step) {
     case CONNECTION_STEP_FIRST:
         /* Fall down to next step */
@@ -591,6 +673,10 @@ connection_step (ConnectionContext *ctx)
         mm_info ("Simple connect state (%d/%d): Connect",
                  ctx->step, CONNECTION_STEP_LAST);
 
+        /* At this point, we can cleanup the cancellation point in the Simple interface,
+         * because the bearer connection has its own cancellation setup. */
+        cleanup_cancellation (ctx);
+
         /* Wait... if we're already using an existing bearer, we need to check if it is
          * already connected; and if so, just don't do anything else */
         if (mm_base_bearer_get_status (ctx->bearer) != MM_BEARER_STATUS_CONNECTED) {
@@ -642,6 +728,12 @@ connect_auth_ready (MMBaseModem *self,
         return;
     }
 
+    if (!setup_cancellation (ctx, &error)) {
+        g_dbus_method_invocation_take_error (ctx->invocation, error);
+        connection_context_free (ctx);
+        return;
+    }
+
     /* We may be able to skip some steps, so check that before doing anything */
     g_object_get (self,
                   MM_IFACE_MODEM_STATE, &current,
@@ -685,8 +777,6 @@ connect_auth_ready (MMBaseModem *self,
 
         mm_dbg ("   Password: %s", VALIDATE_UNSPECIFIED (mm_simple_connect_properties_get_password (ctx->properties)));
 
-        mm_dbg ("   Number: %s", VALIDATE_UNSPECIFIED (mm_simple_connect_properties_get_number (ctx->properties)));
-
 #undef VALIDATE_UNSPECIFIED
     }
 
@@ -741,6 +831,8 @@ handle_connect (MmGdbusModemSimple *skeleton,
     ctx->self = g_object_ref (self);
     ctx->dictionary = g_variant_ref (dictionary);
 
+    mm_dbg ("User request to connect modem");
+
     mm_base_modem_authorize (MM_BASE_MODEM (self),
                              invocation,
                              MM_AUTHORIZATION_DEVICE_CONTROL,
@@ -829,6 +921,7 @@ disconnect_auth_ready (MMBaseModem *self,
 {
     GError *error = NULL;
     MMBearerList *list = NULL;
+    Private *priv;
 
     if (!mm_base_modem_authorize_finish (self, res, &error)) {
         g_dbus_method_invocation_take_error (ctx->invocation, error);
@@ -836,6 +929,14 @@ disconnect_auth_ready (MMBaseModem *self,
         return;
     }
 
+    /* If not disconnecting a specific bearer, also cancel any ongoing
+     * connection attempt. */
+    priv = get_private (MM_IFACE_MODEM_SIMPLE (self));
+    if (!ctx->bearer_path && priv->ongoing_connect) {
+        g_cancellable_cancel (priv->ongoing_connect);
+        g_clear_object (&priv->ongoing_connect);
+    }
+
     g_object_get (self,
                   MM_IFACE_MODEM_BEARER_LIST, &list,
                   NULL);
@@ -883,11 +984,18 @@ handle_disconnect (MmGdbusModemSimple *skeleton,
     ctx->self = g_object_ref (self);
     ctx->invocation = g_object_ref (invocation);
 
-    if (bearer_path &&
-        bearer_path[0] == '/' &&
-        bearer_path[1]) {
-        ctx->bearer_path = g_strdup (ctx->bearer_path);
-    }
+    /* The Disconnect() method expects a valid object path given in bearer path,
+     * it cannot be NULL, so we assume we get '/' when we're asked to disconnect
+     * all connected bearers, as that is what mm_modem_simple_disconnect() does
+     * when a NULL bearer path is given to that method.
+     *
+     * We will detect the '/' string and set the bearer path as NULL in the
+     * context if so, and otherwise use the given input string as path */
+    if (g_strcmp0 (bearer_path, "/") != 0) {
+        mm_dbg ("User request to disconnect modem (bearer '%s')", bearer_path);
+        ctx->bearer_path = g_strdup (bearer_path);
+    } else
+        mm_dbg ("User request to disconnect modem (all bearers)");
 
     mm_base_modem_authorize (MM_BASE_MODEM (self),
                              invocation,
