@@ -637,7 +637,7 @@ port_context_run (MMPluginManager     *self,
      * cancellation error, regardless of what the standard logic does. */
     port_context->cancellable = g_cancellable_new ();
 
-    /* Create a inner task for the port context. The result we expect is the
+    /* Create an inner task for the port context. The result we expect is the
      * best plugin found for the port. */
     port_context->task = g_task_new (self, port_context->cancellable, callback, user_data);
 
@@ -1308,7 +1308,7 @@ device_context_run (MMPluginManager     *self,
      * cancellation error, regardless of what the standard logic does. */
     device_context->cancellable = g_cancellable_new ();
 
-    /* Create a inner task for the device context. We'll complete this task when
+    /* Create an inner task for the device context. We'll complete this task when
      * the last port has been probed. */
     device_context->task = g_task_new (self, device_context->cancellable, callback, user_data);
 }
@@ -1496,6 +1496,36 @@ mm_plugin_manager_peek_plugin (MMPluginManager *self,
 
 /*****************************************************************************/
 
+static void
+register_plugin_whitelist_tags (MMPluginManager *self,
+                                MMPlugin        *plugin)
+{
+    const gchar **tags;
+    guint         i;
+
+    if (!mm_filter_check_rule_enabled (self->priv->filter, MM_FILTER_RULE_PLUGIN_WHITELIST))
+        return;
+
+    tags = mm_plugin_get_allowed_udev_tags (plugin);
+    for (i = 0; tags && tags[i]; i++)
+        mm_filter_register_plugin_whitelist_tag (self->priv->filter, tags[i]);
+}
+
+static void
+register_plugin_whitelist_product_ids (MMPluginManager *self,
+                                       MMPlugin        *plugin)
+{
+    const mm_uint16_pair *product_ids;
+    guint                 i;
+
+    if (!mm_filter_check_rule_enabled (self->priv->filter, MM_FILTER_RULE_PLUGIN_WHITELIST))
+        return;
+
+    product_ids = mm_plugin_get_allowed_product_ids (plugin);
+    for (i = 0; product_ids && product_ids[i].l; i++)
+        mm_filter_register_plugin_whitelist_product_id (self->priv->filter, product_ids[i].l, product_ids[i].r);
+}
+
 static MMPlugin *
 load_plugin (const gchar *path)
 {
@@ -1609,6 +1639,10 @@ load_plugins (MMPluginManager *self,
         else
             /* Vendor specific plugin */
             self->priv->plugins = g_list_append (self->priv->plugins, plugin);
+
+        /* Register plugin whitelist rules in filter, if any */
+        register_plugin_whitelist_tags (self, plugin);
+        register_plugin_whitelist_product_ids (self, plugin);
     }
 
     /* Check the generic plugin once all looped */
